@@ -1,12 +1,13 @@
 package Dancer2::Plugin::Auth::Extensible;
 
-our $VERSION = '0.601';
+our $VERSION = '0.602';
 
 use strict;
 use warnings;
 use Carp;
 use Dancer2::Core::Types qw(Bool HashRef Str);
 use Module::Runtime qw(use_module);
+use Scalar::Util;
 use Session::Token;
 use Try::Tiny;
 use Dancer2::Plugin;
@@ -198,6 +199,8 @@ sub BUILD {
         my $login_page  = $plugin->login_page;
         my $denied_page = $plugin->denied_page;
 
+        Scalar::Util::weaken( my $weak_plugin = $plugin );
+
         # Match optional reset code, but not "denied"
         $app->add_route(
             method => 'get',
@@ -205,16 +208,16 @@ sub BUILD {
             code   => sub {
                 my $app = shift;
 
-                if ( $plugin->logged_in_user ) {
+                if ( $weak_plugin->logged_in_user ) {
                     $app->redirect( $app->request->params->{return_url}
-                          || $plugin->user_home_page );
+                          || $weak_plugin->user_home_page );
                 }
 
                 # Reset password code submitted?
                 my ($code) = $app->request->splat;
 
-                if (   $plugin->reset_password_handler
-                    && $plugin->user_password(code => $code ) )
+                if (   $weak_plugin->reset_password_handler
+                    && $weak_plugin->user_password(code => $code ) )
                 {
                     $app->request->params->{password_code_valid} = 1;
                 }
@@ -223,7 +226,7 @@ sub BUILD {
                 }
 
                 no strict 'refs';
-                return &{ $plugin->login_page_handler }($plugin);
+                return &{ $weak_plugin->login_page_handler }($weak_plugin);
             },
         );
 
@@ -234,7 +237,7 @@ sub BUILD {
                 my $app = shift;
                 $app->response->status(403);
                 no strict 'refs';
-                return &{ $plugin->permission_denied_page_handler }($plugin);
+                return &{ $weak_plugin->permission_denied_page_handler }($weak_plugin);
             },
         );
     }
@@ -506,9 +509,6 @@ sub update_current_user {
 sub update_user {
     my ( $plugin, $username, %update ) = @_;
 
-    my $request = $plugin->app->request;
-    my $session = $plugin->app->session;
-
     my @all_realms = keys %{ $plugin->realms };
     croak "Realm must be specified when more than one realm configured"
       if !$update{realm} && @all_realms > 1;
@@ -516,8 +516,8 @@ sub update_user {
     my $realm    = delete $update{realm} || $all_realms[0];
     my $provider = $plugin->auth_provider($realm);
     my $updated  = $provider->set_user_details( $username, %update );
-    my $cur_user = $session->read('logged_in_user');
-    $request->vars->{logged_in_user_hash} = $updated
+    my $cur_user = $plugin->app->session->read('logged_in_user');
+    $plugin->app->request->vars->{logged_in_user_hash} = $updated
       if $cur_user && $cur_user eq $username;
     $updated;
 }
@@ -525,14 +525,12 @@ sub update_user {
 sub user_has_role {
     my $plugin = shift;
 
-    my $session = $plugin->app->session;
-
     my ( $username, $want_role );
     if ( @_ == 2 ) {
         ( $username, $want_role ) = @_;
     }
     else {
-        $username  = $session->read('logged_in_user');
+        $username  = $plugin->app->session->read('logged_in_user');
         $want_role = shift;
     }
 
@@ -632,9 +630,9 @@ sub user_password {
 
 sub user_roles {
     my ( $plugin, $username, $realm ) = @_;
-    my $session = $plugin->app->session;
 
-    $username = $session->read('logged_in_user') unless defined $username;
+    $username = $plugin->app->session->read('logged_in_user')
+      unless defined $username;
 
     my $search_realm = ( $realm ? $realm : '' );
 
@@ -982,7 +980,7 @@ sub _post_login_route {
     if ($code) {
         no strict 'refs';
         my $randompw = &{ $plugin->password_generator };
-        if ( $app->user_password( code => $code, new_password => $randompw ) ) {
+        if ( $plugin->user_password( code => $code, new_password => $randompw ) ) {
             $app->forward(
                 $plugin->login_page,
                 { new_password => $randompw },
@@ -1022,8 +1020,13 @@ sub _post_login_route {
       $plugin->authenticate_user( $username, $password, $auth_realm );
 
     if ($success) {
-        $app->app->session->write( logged_in_user       => $username );
-        $app->app->session->write( logged_in_user_realm => $realm );
+
+        # change session ID if we have a new enough D2 version with support
+        $plugin->app->change_session_id
+          if $plugin->app->can('change_session_id');
+
+        $app->session->write( logged_in_user       => $username );
+        $app->session->write( logged_in_user_realm => $realm );
         $app->log( core => "Realm is $realm" );
         $app->redirect( $app->request->params->{return_url}
               || $plugin->user_home_page );
@@ -1765,7 +1768,7 @@ Documentation fix by Vince Willems.
 
 Henk van Oers (GH #8, #13).
 
-Andrew Beverly (GH #6, #7, #10, #17, #22, #24, #25, #26).
+Andrew Beverly (GH #6, #7, #10, #17, #22, #24, #25, #26, #53).
 This includes support for creating and editing users and manage user passwords.
 
 Gabor Szabo (GH #11, #16, #18).
